// <cuckoo-string.cpp> -*- C++ -*-

#include <stdlib.h>
#include <iostream>
#include <vector>
#include <math.h>
#include <string>

#include <sys/times.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <set>

// #define USE_PREFETCH

#ifdef USE_PREFETCH
#define PREFETCH( x ) __builtin_prefetch( x )
#define PREFETCH_RW( x ) __builtin_prefetch( x, 1 )
#else
#define PREFETCH( x )
#define PREFETCH_RW( x )
#endif

using namespace std;

template <class K>
class HashFun{
protected:
  unsigned long m_nBuckets;
public:
  HashFun( const unsigned long nBuckets = 1 ){
    m_nBuckets = nBuckets;
  };
  ~HashFun( ){};
  typedef K Key;
  typedef long HashVal;
};

#define MAX_STRLEN 4096

/** Diese Klasse erzeugt eine groe Tabelle mit Zufallszahlen */
static class RANDOM_TABLE{
protected:
  unsigned int table[256*MAX_STRLEN*2];
public:
  RANDOM_TABLE( ){
    for( int j = 256*MAX_STRLEN*2-1; j>= 0; j-- )
      table[j] = lrand48( );
  }
  const unsigned int *get_addr( int j ){
    return &table[j];
  }
} random_table;




/*
  Diese Klasse kapselt einen C-String.
  Um die Cache-Effizienz bei Vergleichen zu steigern, werden zu jedem
  Index eines Strings auch dessen Prfsumme gespeichert.
*/

class checksum_entry{
protected:
  int finger;
  const char *p;
public:
  checksum_entry( const char *_p = NULL, const int _finger = -1 ) : 
    p( _p ), finger( _finger ) { }

  operator const char*() const {
    return p;
  }

  const char* get_p( ) const {
    return p;
  }

  int get_finger( ) const { return finger; };
} __attribute__ ((packed));

inline bool operator ==( const checksum_entry &a, const checksum_entry &b ){
  if( a.get_finger( ) != b.get_finger( ) ) return false;
  if( a.get_p( ) == b.get_p( ) ) return true;
  return strcmp( a.get_p( ), b.get_p( ) ) == 0;
}

inline bool operator !=( const checksum_entry &a, const checksum_entry &b ){
  if( a.get_finger( ) != b.get_finger( ) ) return true;
  if( a.get_p( ) == b.get_p( ) ) return false;
  return strcmp( a.get_p( ), b.get_p( ) ) != 0;
}

inline bool is_empty ( const checksum_entry &a ){
  return a.get_finger( ) < 0;
}


typedef const char *PCHAR;
struct PSTRING {
  PCHAR p;
  operator const char*() const {
    return p;
  }
  const char* get_p( ) const {
    return p;
  }

  int get_finger( ) const { return 0; };
  
  

  PSTRING( PCHAR x = NULL, int _finger = 0 ) : p(x){};
};




inline bool operator ==( const PSTRING &a, const PSTRING &b ){
  return strcmp( a.p, b.p ) == 0;
}

inline bool operator !=( const PSTRING &a, const PSTRING &b ){
  return strcmp( a.p, b.p ) != 0;
}

inline bool is_empty ( const PSTRING &a ){
  return a.p == NULL;
}


/** Eine zufllig gewhlte 
 * Carter-Wegmann-Hashfunktion fr Strings. */
class stringStaticHashFun_xor : public HashFun<PCHAR>{
protected:
  const unsigned int *m_table;
public:  
  stringStaticHashFun_xor( const unsigned long nBuckets = 1 ) : HashFun<Key>( nBuckets ){
    m_table = random_table.get_addr( lrand48( ) % ( 256 * MAX_STRLEN ) );
  }
  HashVal operator()( const Key &s ) const {
    unsigned int hv = m_table[0], index=0;
    const char* s_ = s;
    char last;
    while( (last = *s_++) != '\0' ){
      hv ^= m_table[ index += last ];
    }
    return hv % m_nBuckets;
  }
};

/** Eine zufllig gewhlte Hashfunktion fr Strings
 * mit einfachen Polynomem */
class stringStaticHashFun_xor_smalltable : public HashFun<PCHAR>{
protected:
  const unsigned int *m_table;
public:  
  stringStaticHashFun_xor_smalltable( const unsigned long nBuckets = 1 ) : HashFun<Key>( nBuckets ){
    m_table = random_table.get_addr( lrand48( ) % ( 256 * MAX_STRLEN ) );
  }
  HashVal operator()( const Key &s ) const {
    unsigned int hv = m_table[0], index=1;
    const char* s_ = s;
    char last;
    while( (last = *s_++) != '\0' ){
      hv ^= m_table[ index++ ] * last;
    }
    return hv % m_nBuckets;
  }
};


/** Eine zufllig gewhlte Hashfunktion fr Strings
 * mit einem einfachen Zufallszahlengenerator */
class stringStaticHashFun_rng_simple : public HashFun<PCHAR>{
protected:
  unsigned m_a, m_b;

public:  
  stringStaticHashFun_rng_simple( const unsigned long nBuckets = 1 ) : 
    HashFun<Key>( nBuckets ) {
    m_a = lrand48( );
    m_b = lrand48( );
  }
  HashVal operator()( const Key &s ) const {
    const char* v = s;
    unsigned int M1 = m_nBuckets-1, h = 0, a = m_a, b = m_b;
    
    for( ; *v != 0; v++, a = a*b % M1 )
      h = (a*h + *v) % m_nBuckets;
    return h % m_nBuckets;
  }
};



/* Ein Paar h0, h1 von Hashfunktionen */
template <class HF>
class f_pair{
public:
  HF h0, h1;
  
  f_pair( const unsigned long nBuckets = 1 ) : h0(nBuckets), h1(nBuckets) {};
  
  void make_pair( const PCHAR &x, int &i0, int &i1 ) const {
    i0 = h0(x);
    i1 = h1(x);
  };
};

/* Ein Paar h, q von Hashfunktionen mit 
 * Trick von R. Pagh: h0(x) = q(x) - h1(x)*/
template <class HF>
class f_pair_improved{
public:
  HF h, q;
  int m_nBuckets;
  
  f_pair_improved( const unsigned long nBuckets = 1 ) : h(nBuckets), q(nBuckets) {
    m_nBuckets = nBuckets;
  };
  
  /** Berechnet q(x) */
  int get_q( const PCHAR &x ) const{
    return q(x);
  }
  
  void make_pair( const PCHAR &x, int &i0, int &i1 ) const {
    i0 = h(x);
    int _q = q(x);
    i1 = _q - i0;
    if( i1 < 0 ) i1 += m_nBuckets;
  };
};


/* Kapselt ein paar wichtige Informationen
 * ber das Wrterbuch. F_PAIR ist Platzhalter fr
 * einen von den Funktionspaar-Datentypen weiter unten.
 */
template <class F_PAIR>
class cuckoo_cache_block_data{
public:
  int d;
  int r; //! Maximale Anzahl von Pltzen im Dictionary
  int nkeys;
  F_PAIR h; /* Das hashfunktionenpaar */
  double resizing_factor; //! Um diesen Faktor wird die Tabelle beim Rehash vergrert
  double epsilon;
  int nMaxTests;
  
  /** Berechnet die Gre des Feldes, wenn _r Schlssel
   * abgescpeichert werden sollen. Diese Funktion kann benutzt werden,
   * wenn man zum Beispiel sicher stellen will, dass die Gre des Feldes
   * ein wenig lnger als _r ist, um den Test auf das Feldende zu
   * vermeiden. */
  int calculate_field_size( const int _r ){
    return (int) (ceil( double(_r) / this->d ) * this->d);
  }
  
  /** Gibt die aktuelle Zahl der Schlssel im Wrterbuch zurck */
  int get_nkeys( ) const { return nkeys; };

  /** Gibt das Verhltnis von Platz zu Schlsselzahl zurck */
  double get_loadfactor( ) const { return double(r) / nkeys; };

  /** Gibt Setzt den Vergrerungsfaktor */
  void set_resizingfactor( const double &_f ) { resizing_factor = _f; };
};


/* Implementiert das Finden des alternativen Blockes
 * bei Verwendung von Fingerabdrcken */
template <class ENTRY, class F_PAIR>
class alternate_finger {
  const cuckoo_cache_block_data<F_PAIR> *pccbd;
public:
  alternate_finger( const cuckoo_cache_block_data<F_PAIR> *_pccbd ) : pccbd(_pccbd){};
  int alternate( const int block_start, const ENTRY &entry ) const {
    /* Es muss keine Hashfunktion ausgewertet werden! */
    return entry.get_finger( ) - block_start;
  }
};

/* Implementiert das Finden des alternativen Blockes
 * ohne Verwendung von Fingerabdrcken */
template <class ENTRY, class F_PAIR>
class alternate_string {
  const cuckoo_cache_block_data<F_PAIR> *pccbd;
public:
  alternate_string( const cuckoo_cache_block_data<F_PAIR> *_pccbd ) : pccbd(_pccbd){};
  int alternate( const int block_start, const ENTRY &entry ) const {
    int i0 = pccbd->h.h0(entry.get_p( )) * pccbd->d;
    return i0 != block_start ? i0 : pccbd->h.h1(entry.get_p( )) * pccbd->d;
  }
};

/* Implementiert das Finden des alternativen Blockes
 * ohne Verwendung von Fingerabdrcken mit Trick von R. Pagh */
template <class ENTRY, class F_PAIR>
class alternate_string_improved {
  const cuckoo_cache_block_data<F_PAIR> *pccbd;
public:
  alternate_string_improved( const cuckoo_cache_block_data<F_PAIR> *_pccbd ) : pccbd(_pccbd){};
  int alternate( const int block_start, const ENTRY &entry ) const {
    int res = pccbd->h.get_q(entry.get_p()) * pccbd->d - block_start;
    if( res < 0 ) res += pccbd->r;
    return res; 
  }
};


/* Cuckoo Hashing */
template<class DICT_ENTRY, class F_PAIR, class ALTERNATE>
class cuckoo_cache_block : public cuckoo_cache_block_data<F_PAIR> {
protected:
  DICT_ENTRY *field; /* Die Daten */
  const DICT_ENTRY empty;
  const ALTERNATE m_alternate;
    
  /** Legt ein Feld mit _r Pltzen fr Schlssel an.
   * Benutzt calculate_field_size, um die Gre des Feldes zu
   * berechnen und setzt r */
  DICT_ENTRY *make_field( const int _r ){
    DICT_ENTRY *f = new DICT_ENTRY[_r];
    for( int i = _r-1; i >= 0; i-- ) 
      f[i] = empty;
    return f;
  }
  
  void make_h( ){
    this->h = F_PAIR( this->r/this->d );
    this->nMaxTests = this->r;
  }

public:
  typedef DICT_ENTRY entry_type;

  cuckoo_cache_block( const double _epsilon = 0.2,
		      const int _d = 16,
		      const int _nkeys = -1 ) : m_alternate(this), empty( ) {
    this->resizing_factor = 2;
    this->d = _d;
    this->epsilon = _epsilon;
    this->r = _nkeys == -1 ? 2*_d : this->calculate_field_size( (int) ((1+_epsilon)*_nkeys) );
    this->nkeys = 0;
    field = make_field( this->r );
    this->make_h( );
 }

  ~cuckoo_cache_block( ){
    delete[] field;
  }

  /** Fhrt einen kompletten Rehash des Wrterbuchs mit Gre _r durch */
  void rehash( const int _r ){
    DICT_ENTRY *old_field = field;
    clog << "[" << _r << flush;
    int old_size = this->rr;
    this->r = this->calculate_field_size( _r );
    field = make_field( this->r );
    this->make_h( );
    this->nkeys = 0;
    for( int i = old_size-1; i >= 0; i-- )
      if( !is_empty( old_field[i] ) ) insert( old_field[i].get_p( ) );
    delete old_field;
    clog << "] " << flush;
  }


  /** Fgt den Schlssel key ins Wrterbuch ein und
   * gibt true zurck, falls das Einfgen erfolgreich war. */
  int insert( const char *px ){
    int i1, i0;
    this->h.make_pair( px, i1, i0 );
    i0 *= this->d;
    PREFETCH_RW( &field[i0] );
    i1 *= this->d;
    PREFETCH_RW( &field[i1] );
    int z = lrand48( ) & 0x80;
    DICT_ENTRY x( px, i1+i0 );
    if( z )
      swap( i1, i0 );
    int free_cell = -1;
    int _i0 = i0;

    int ende = i0+this->d;
    bool bempty1 = false, bempty0 = false;
    for( ; i0 < ende; i0++, i1++ ){
      bempty0 = is_empty( field[i0] );
      bempty1 = is_empty( field[i1] );
      if( !bempty0 ){
        if( field[i0] == x )
          return i0;
      }
      else if( free_cell == -1 )
        free_cell = i0;
      if( !bempty1 ){
        if( field[i1] == x )
          return i1;
      }
      else if( free_cell == -1 )
        free_cell = i1;
      if( bempty0 && bempty1 )
        break;
    }

    if( free_cell != -1 ){
      field[ free_cell ] = x;
      this->nkeys++;
      return free_cell;
    }

    int counter = this->nMaxTests;
    while( counter-- ){
      int j = _i0 + lrand48( ) % this->d;
      swap( field[j], x );
      _i0 = m_alternate.alternate( _i0, x );
      
      // Suche freies Feld
      ende = _i0+this->d;
      int offset = 0;
      for( int i = _i0; i < ende; i++ ){
        if( is_empty( field[i] ) ){
          field[i] = x;
          this->nkeys++;
          return i;
        }
      }
    }
    return -1;
    /* Anstatt das Einfgen hier zu beenden, knnte man folgendes tun,
     * um ein richtiges dynamisches Wrterbuch zu bekommen:
  
     rehash( max( r+1, (int)(resizing_factor*r) ) );
     return insert( x.get_p( ) );
    */
  }


  /** Gibt die Adresse des Schlssels key zurck oder -1, falls
   * key nicht im Wrterbuch drin ist. */
  int lookup( const char *px ) const{
    int i1, i0;
    this->h.make_pair( px, i1, i0 );
    i1 *= this->d;
    PREFETCH( &field[i1] );
    i0 *= this->d;
    PREFETCH( &field[i0] );
    DICT_ENTRY x( px, i1+i0 );

    bool bempty0 = is_empty( field[i0] );
    if( !bempty0 && field[i0] == x )
      return i0;
    i0++;
    bool bempty1 = is_empty( field[i1] );
    if( !bempty1 && field[i1] == x )
      return i1;
    int ende=i0+this->d;
    i1++;
    for( ; i0 < ende; i0++, i1++ ){
      if( !bempty0 ){
        bempty0 = is_empty( field[i0] );
        if( !bempty0 && field[i0] == x )
          return i0;
      }
      if( !bempty1 ){
        bempty1 = is_empty( field[i1] );
        if( !bempty1 && field[i1] == x )
          return i1;
      }
    }
    return -1;
  }


  /** Lscht den Schlssel key und gibt true zurck, falls
   * das Lschen erfolgreich war */
  bool delete_key( const char *px ){
    int zelle;

    zelle = -1;
    int i1, i0;
    this->h.make_pair(px, i0, i1);
    i1 *= this->d;
    PREFETCH_RW( &field[i1] );
    i0 *= this->d;
    PREFETCH_RW( &field[i0] );
    DICT_ENTRY x( px, i1+i0 );
    int block_ende0 = i0+this->d;
    int block_ende1 = i1+this->d;
    int ende;
    for( ; i0 < block_ende0; i0++, i1++ ){
      if( !is_empty( field[i0] ) && field[i0] == x ){
        ende = block_ende0;
        zelle = i0;
        break;
      }
      if( !is_empty( field[i1] ) && field[i1] == x ){
        ende = block_ende1;
        zelle = i1;
        break;
      }
    }
    if( zelle == -1 )
      return false;
    int _zelle = zelle;
    /* Wir lschen das Element und schlieen die Lcke, die x hinterlsst */
    zelle++;
    while( zelle < ende && !is_empty( field[zelle] ) ) zelle++;
    zelle--; // Jetzt zeigte Zelle auf den letzten belegten Index

    if( zelle > _zelle )
      field[_zelle] = field[zelle];
    field[zelle] = empty;
    this->nkeys--;
    return true;
  }

  /** Zhlt die Schlssel im Wrterbuch zurck */
  int count_keys( ) const{
    int c = 0;
    for( int i = this->r-1; i >= 0; i-- )
      if( !is_empty( field[i] ) ) c++;
    return c;
  }


  /** Gibt den Schlssel mit dem Index j zurck */
  const DICT_ENTRY &key( int j ) const {
    return field[ j ];
  }
};



/** Hlt eine Liste der Zeiger auf die Strings.
 * Um die amortisierte Laufzeit niedrig zu halten, wird die Verdoppelungs-strategie verwendet.
 */
class persistent_stringarray : public vector<PCHAR>{
protected:
  bool m_bFreeOnDelete;

public:
  persistent_stringarray( bool bFreeOnDelete = true ) { m_bFreeOnDelete = bFreeOnDelete; };
  void set_freeOnDelete( bool bFreeOnDelete ){ m_bFreeOnDelete = bFreeOnDelete; };
  
  /** Fgt einen String ans Ende ein und gibt dessen Index zurck */
  int insert( const PCHAR &s ){
    if( size( ) >= capacity( ) ){
      reserve( 2*size( ) );
    }
    int ind = size( );
    push_back( s );
    return ind;
  }
  
  /** Fgt eine Kopie ans Ende ein und gibt dessen Index zurck */
  int insert_copy( const PCHAR &s ){
    int len = strlen( s );
    PCHAR s_copy = (PCHAR)malloc( len+1 );
    memcpy( (void*)s_copy, s, len+1 );
    return insert( s_copy );
  }

  ~persistent_stringarray( ){
    if( m_bFreeOnDelete )
      for( iterator it = begin( ); it != end( ); it++ )
        free( (void*) *it );
  }
};







tms tms_start, tms_stop;

inline void start( ){
  times( &tms_start );
};

inline clock_t stop( ){
  times( &tms_stop );
  return tms_stop.tms_stime + tms_stop.tms_utime - tms_start.tms_stime - tms_start.tms_utime;
};



double the_epsilon = 1;
int verfahren = 0;
int the_d = 8;
template< class DICT >
void make_tests( ){
  string s;
  time_t t;
  srand48( time( &t ) );
  
  /* Erstmal alle Zeilen einlesen */
  clog << "Lese Eingabe... " << flush;
  int cnt = 0;
  start( );
  persistent_stringarray all_strings(false), all_neg_strings(true);
  while (getline(cin,s,'\n')){
    int index = all_neg_strings.insert_copy( ('@' + s).c_str( ) );
    all_strings.insert( all_neg_strings[ index ] + 1 );
    cnt++;
    if( cnt % 0x10000 == 0 )
      clog << "." << flush;
  }
  t = stop( );
  double et = double(t) / sysconf( _SC_CLK_TCK );
  clog << all_strings.size( ) << " Zeichenketten gelesen." << endl;
  cout << "Zeit zum Einlesen: " << et << " Sekunden" << endl;

  DICT dict( the_epsilon, the_d, all_strings.size( ) );
  dict.set_resizingfactor(1.5);
  clog << "Gre eines Eintrages im Wrterbuch: " << sizeof( dict.key( 0 ) ) << endl;
  cnt = 0;
  clog << "Fge ein... " << flush;
  start( );
  for( int j = all_strings.size( )-1; j >= 0; j-- ){
    if( dict.insert( all_strings[j] ) == -1 ){
      cout << "Build: ***" << endl;
      exit( -1 );
    }
  }
  clog << "Fertig" << endl << flush;
  t = stop( );
  et = double(t) / sysconf( _SC_CLK_TCK );
  cout << "Build: " << 1000000*et/all_strings.size( ) << " s" << endl;
  
  clog << "count_keys = " << dict.count_keys( ) << endl;
  clog << "get_nkeys = " << dict.get_nkeys( ) << endl;
  cout << "Auslastung des Wrterbuchs: " << 0.1*floor(1000/dict.get_loadfactor( )) << "%" << endl;
 
  /* Tests */
  clog << "Postitive lookups... " << flush;
  start( );
  for( int j = all_strings.size( )-1; j >= 0; j-- ){
    int index = dict.lookup( all_strings[j] );
    if( index == -1 ){
      cerr << "Schlssel '" << all_strings[j] << "' nicht gefunden. Das geht so nicht..." << endl;
      exit( -1 );
    } else if( strcmp( dict.key( index ), all_strings[j] ) != 0 ){
      cerr << "Da stimmen zwei Schlssel nicht. Das geht so nicht..." << endl;
      exit( -1 );
   }
  }
  t = stop( );
  clog << "Fertig" << endl << flush;
  et = double(t) / sysconf( _SC_CLK_TCK );
  cout << "Pos. Lookups: " << 1000000*et/all_strings.size( ) << " s" << endl;
  
  clog << "Negative lookups... " << flush;
  start( );
  for( int j = all_neg_strings.size( )-1; j >= 0; j-- ){
    int index = dict.lookup( all_neg_strings[j] );
    if( index != -1 ){
      cerr << "Schlssel '" << all_neg_strings[j] << "' nicht gefunden. Das geht so nicht..." << endl;
      exit( -1 );
    } 
  }
  t = stop( );
  clog << "Fertig" << endl << flush;
  et = double(t) / sysconf( _SC_CLK_TCK );
  cout << "Neg. Lookups: " << 1000000*et/all_strings.size( ) << " s" << endl;

  clog << "remove/insert... " << flush;
  start( );
  for( int j = all_strings.size( )-1; j >= 0; j-- ){
    dict.delete_key( all_strings[j] );
    if( dict.insert( all_neg_strings[j] ) == -1 ){
      cout << "remove/insert: ***" << endl;
      clog << "count_keys = " << dict.count_keys( ) << endl;
      clog << "get_nkeys = " << dict.get_nkeys( ) << endl;
      exit( -1 );
    }
  }
  t = stop( );
  clog << "Fertig" << endl << flush;
  et = double(t) / sysconf( _SC_CLK_TCK );
  cout << "remove/insert: " << 1000000*et/all_strings.size( ) << " s" << endl;
}



typedef f_pair<stringStaticHashFun_xor> xor_pair;
typedef f_pair_improved<stringStaticHashFun_xor> xor_pair_improved;
typedef f_pair<stringStaticHashFun_rng_simple> rng_pair;
typedef f_pair_improved<stringStaticHashFun_rng_simple> rng_pair_improved;
typedef f_pair<stringStaticHashFun_xor_smalltable> smalltable_pair;
typedef f_pair_improved<stringStaticHashFun_xor_smalltable> smalltable_pair_improved;


template< class DICT >
class auto_test{
public:
  int nRunden;
  vector<double> all_alpha;
  set<int> all_d;
  persistent_stringarray all_strings, all_neg_strings;
  
  tms tms_start, tms_stop;
  time_t t;
  
  inline void start( ){
    times( &tms_start );
  };
  
  inline clock_t stop( ){
    times( &tms_stop );
    return tms_stop.tms_stime + tms_stop.tms_utime 
      - tms_start.tms_stime - tms_start.tms_utime;
  };
  
  auto_test( const int _nRunden = 5 ) : all_strings(false), all_neg_strings(true) {
    nRunden = _nRunden;
    for( int d=16; d > 0; d-- )
      all_d.insert( d ); 
    for( double alpha=0.8; alpha >= 0.301; alpha-=0.01 )
      all_alpha.push_back( alpha );
    read_strings( );
    srand48( time( &t ) );
  }
  
  void read_strings( ){
    /* Erstmal alle Zeilen einlesen */
    clog << "Lese Eingabe... " << flush;
    int cnt = 0;
    string s;
    start( );
    while (getline(cin,s,'\n')){
      int index = all_neg_strings.insert_copy( ('@' + s).c_str( ) );
    all_strings.insert( all_neg_strings[ index ] + 1 );
    cnt++;
    if( cnt % 0x10000 == 0 )
      clog << "." << flush;
    }
    t = stop( );
    double et = double(t) / sysconf( _SC_CLK_TCK );
    clog << all_strings.size( ) << " Zeichenketten gelesen." << endl;
    clog << "Zeit zum Einlesen: " << et << " Sekunden" << endl;
  };

  struct STAT{
    int build_ok, delins_ok;
    double build_s, pos_lkp_s, neg_lkp_s, delins_s;
    STAT( ){
      build_ok = delins_ok = 0;
      build_s = pos_lkp_s = neg_lkp_s = delins_s = 0;
    }
  };

  void one_auto_test( const double &alpha ){
    /* Hier wird das zu alpha passende epsilon berechnet. Die Formel steht in der Arbeit */
    double the_epsilon = sizeof( PCHAR ) / ( alpha * sizeof( typename DICT::entry_type ) ) - 1;
    clog << "Teste fr alpha = " << alpha << " [epsilon = " << the_epsilon << "]" << endl;
    if( the_epsilon <= 0.0001 )
      return;
    for( set<int>::const_reverse_iterator d = all_d.rbegin( ); d != all_d.rend( ); d++ ){
      clog << "Beginne mit d = " << *d << ". Verbleibende Runden: " << flush;
      STAT stat;
      for( int k = nRunden; k>0; k-- ){
	clog << k << " " << flush;
	DICT dict( the_epsilon, *d, all_strings.size( ) );
	dict.set_resizingfactor(1.5);
	bool build_ok = true;
	start( );
	for( int j = all_strings.size( )-1; j >= 0; j-- ){
	  if( dict.insert( all_strings[j] ) == -1 ){
	    build_ok = false;
	    clog << "[Build ***] " << flush;
	    break;
	  }
	}
	t = stop( );
	if( build_ok ){
	  stat.build_ok++;
	  stat.build_s+= double(t) / sysconf( _SC_CLK_TCK );
	  
	  /* Tests */
	  start( );
	  for( int j = all_strings.size( )-1; j >= 0; j-- ){
	    int index = dict.lookup( all_strings[j] );
	    if( index == -1 ){
	      cerr << "Schlssel '" << all_strings[j] << "' nicht gefunden. Das geht so nicht..." << endl;
	      exit( -1 );
	    } else if( strcmp( dict.key( index ), all_strings[j] ) != 0 ){
	      cerr << "Da stimmen zwei Schlssel nicht. Das geht so nicht..." << endl;
	      exit( -1 );
	    }
	  }
	  t = stop( );
	  stat.pos_lkp_s+=double(t) / sysconf( _SC_CLK_TCK );
	  
	  start( );
	  for( int j = all_neg_strings.size( )-1; j >= 0; j-- ){
	    int index = dict.lookup( all_neg_strings[j] );
	    if( index != -1 ){
	      cerr << "Schlssel '" << all_neg_strings[j] << "' nicht gefunden. Das geht so nicht..." << endl;
	      exit( -1 );
	    } 
	  }
	  t = stop( );
	  stat.neg_lkp_s+= double(t) / sysconf( _SC_CLK_TCK );
	  
	  bool delins_ok = true;
	  start( );
	  for( int j = all_strings.size( )-1; j >= 0; j-- ){
	    dict.delete_key( all_strings[j] );
	    if( dict.insert( all_neg_strings[j] ) == -1 ){
	      clog << "[delins ***] " << endl;
	      delins_ok = false;
	      break;
	    }
	  }
	  t = stop( );
	  if( delins_ok ){
	    stat.delins_ok++;
	    stat.delins_s += double(t) / sysconf( _SC_CLK_TCK );
	  }
	}
      }
      clog << "Fertig" << endl << flush;
      double f = 1000000.0 / all_strings.size( );
      stat.build_s *= f;
      stat.pos_lkp_s *= f;
      stat.neg_lkp_s *= f;
      stat.delins_s  *= f;
      if( stat.build_ok > 0 ){
	stat.build_s /= stat.build_ok;
	stat.pos_lkp_s /= stat.build_ok;
	stat.neg_lkp_s /= stat.build_ok;
      }
      if( stat.delins_ok > 0 ){
	stat.delins_s /= stat.delins_ok;
      }
      clog << "Build: " << stat.build_s << " s" << endl;
      clog << "Pos. Lookups: " << stat.pos_lkp_s << " s" << endl;
      clog << "Neg. Lookups: " << stat.neg_lkp_s << " s" << endl;
      clog << "remove/insert: " << stat.delins_s << " s" << endl;
      cout << alpha << "\t" << *d << "\t";
      if( stat.build_ok > 0 ) 
	cout << stat.build_s; 
      else
	cout << "\\N";
      cout << "\t" << stat.build_ok << "\t";
      if( stat.build_ok > 0 ) 
	cout << stat.pos_lkp_s;
      else
	cout << "\\N";
      cout << "\t";
      if( stat.build_ok > 0 ) 
	cout << stat.neg_lkp_s;
      else
	cout << "\\N";
      cout << "\t";
      if( stat.delins_ok > 0 ) 
	cout << stat.delins_s;
      else
	cout << "\\N";
      cout << "\t" << stat.delins_ok << endl;
      if( 2*stat.build_ok < nRunden )
	break;
    }
  };
  void make_tests( ){
    for( vector<double>::const_iterator alpha = all_alpha.begin( ); alpha != all_alpha.end( ); alpha++ ){
      one_auto_test( *alpha );
    }
  }
};

template< class DICT >
void auto_tests( ){
  auto_test<DICT> at;
  at.make_tests( );
};

void call_auto_tests( int verfahren ){
  switch( verfahren ){
  case 1 : auto_tests< cuckoo_cache_block<checksum_entry, xor_pair, alternate_finger<checksum_entry, xor_pair> > > ( ); break;
  case 2 : auto_tests< cuckoo_cache_block<PSTRING, xor_pair_improved, alternate_string_improved<PSTRING, xor_pair_improved> > > ( ); break;
  case 3 : auto_tests< cuckoo_cache_block<PSTRING, xor_pair, alternate_string<PSTRING, xor_pair> > > ( ); break;
  case 4 : auto_tests< cuckoo_cache_block<checksum_entry, rng_pair, alternate_finger<checksum_entry, rng_pair> > > ( ); break;
  case 5 : auto_tests< cuckoo_cache_block<PSTRING, rng_pair_improved, alternate_string_improved<PSTRING, rng_pair_improved> > > ( ); break;
  case 6 : auto_tests< cuckoo_cache_block<PSTRING, rng_pair, alternate_string<PSTRING, rng_pair> > > ( ); break;
  case 7 : auto_tests< cuckoo_cache_block<checksum_entry, smalltable_pair, alternate_finger<checksum_entry, smalltable_pair> > > ( ); break;
  case 8 : auto_tests< cuckoo_cache_block<PSTRING, smalltable_pair_improved, alternate_string_improved<PSTRING, smalltable_pair_improved> > > ( ); break;
  case 9 : auto_tests< cuckoo_cache_block<PSTRING, smalltable_pair, alternate_string<PSTRING, smalltable_pair> > > ( ); break;
  }
}

void make_simple_test( int verfahren ){
  switch( verfahren ){
  case 1 : make_tests< cuckoo_cache_block<checksum_entry, xor_pair, alternate_finger<checksum_entry, xor_pair> > > ( ); break;
  case 2 : make_tests< cuckoo_cache_block<PSTRING, xor_pair_improved, alternate_string_improved<PSTRING, xor_pair_improved> > > ( ); break;
  case 3 : make_tests< cuckoo_cache_block<PSTRING, xor_pair, alternate_string<PSTRING, xor_pair> > > ( ); break;
  case 4 : make_tests< cuckoo_cache_block<checksum_entry, rng_pair, alternate_finger<checksum_entry, rng_pair> > > ( ); break;
  case 5 : make_tests< cuckoo_cache_block<PSTRING, rng_pair_improved, alternate_string_improved<PSTRING, rng_pair_improved> > > ( ); break;
  case 6 : make_tests< cuckoo_cache_block<PSTRING, rng_pair, alternate_string<PSTRING, rng_pair> > > ( ); break;
  case 7 : make_tests< cuckoo_cache_block<checksum_entry, smalltable_pair, alternate_finger<checksum_entry, smalltable_pair> > > ( ); break;
  case 8 : make_tests< cuckoo_cache_block<PSTRING, smalltable_pair_improved, alternate_string_improved<PSTRING, smalltable_pair_improved> > > ( ); break;
  case 9 : make_tests< cuckoo_cache_block<PSTRING, smalltable_pair, alternate_string<PSTRING, smalltable_pair> > > ( ); break;
  }
}



/** Testprogramm frs dictionary: Von der Standardeingabe
 * werden zeilenweise Strings gelesen und anschlieend ins Wrterbuch eingefgt.
 * Dann werden ein paar Tests durchgefhrt */
int main( int argc, char *argv[] ){
  if( argc < 2 ){
    cout << argv[0] << " verfahren [ epsilon [ d ] ]" << endl;
    cout << "verfahren = 1: CW-Wegmann-Funktion mit Fingerabdrcken" << endl;
    cout << "verfahren = 2: CW-Wegmann-Funktion und verbesserte Hashfunktionen" << endl;
    cout << "verfahren = 3: CW-Wegmann-Funktion" << endl;
    cout << "verfahren = 4: Einfache Polynome mit Fingerabdrcken" << endl;
    cout << "verfahren = 5: Einfache Polynome und verbesserte Hashfunktionen" << endl;
    cout << "verfahren = 6: Einfache Polynome" << endl;
    cout << "verfahren = 7: Einfacher Zufallszahlengenerator mit Fingerabdrcken" << endl;
    cout << "verfahren = 8: Einfacher Zufallszahlengenerator und verbesserte Hashfunktionen" << endl;
    cout << "verfahren = 9: Einfacher Zufallszahlengenerator" << endl;
    cout << "Wird nur verfahren angegeben, wird eine Messreihe fr verschiedene\nalpha und d gestartet." << endl;
    exit( -1 );
  }
  clog << "Gre PSTRING: " << sizeof(PSTRING) << endl;
  clog << "Gre int: " << sizeof(int) << endl;
  if( argc >= 2 ) verfahren = atoi( argv[ 1 ] );
  if( argc >= 3 ){ 
    the_epsilon = atof( argv[ 2 ] );
    if( argc >= 4 ) the_d = atoi( argv[ 3 ] );
    make_simple_test( verfahren );
  }
  else{
    call_auto_tests( verfahren );
  }

  return 0;
}
